/************************************************************************

  	PIC Etherfeed - tickling your ego or statistics? If you have a
	website, blog or any other web service than the etherfeed is for you.
	The etherfeed is a small standalone network device, it connects to the
	Internet and displays statistical data about your website like number
	of unique visitors, number of pageviews, PayPal sales (SUM of sales via
	the PayPal IPN), Facebook likes for your brand / group and weather data
	like temperature, humidity for your location.
	
	OpenSource & OpenHardware project with kits available at:

	http://www.etherfeed.com
	https://www.facebook.com/etherfeed

    Copyright (C) 2012  Boris Ribov, Georgi Bakalski, Militsa Ilieva

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

 *********************************************************************
 *
 *  LCD Access Routines
 * 	FileName: LCD.c
 *
 *********************************************************************
 * Boris Ribov   01/07/2012    Rev. 2.0		; Support reading the busy 
 *											; flag on data pin 7. Adding
 *											; support to some different
 *											; displays using define flags;
 ********************************************************************/
#define __LCD_C
#include "TCPIP Stack/TCPIP.h"

// LCDText is a 80 byte shadow of the LCD text.
// Write to it and then call LCDUpdate() to copy
// the string into the LCD module.
BYTE LCDText[20*4+1];

/******************************************************************************
 * Function:        static void LCDWrite(BYTE RS, BYTE Data)
 * PreCondition:    None
 * Input:           RS - Register Select - 1:RAM, 0:Config registers
 *					Data - 8 bits of data to write
 *
 * Output:          None
 * Side Effects:    None
 * Overview:        Controls the Port I/O pins to cause an LCD write
 * Note:            None
 *****************************************************************************/
static void LCDWrite(BYTE RS, BYTE Data)
{
	// Set IO data or command mode
	LCD_RS = RS;
	Nop ();

	// Set data on the bus
	PORTB = Data;
	Nop ();
	LCD_E = 1;
	Nop ();
	LCD_E = 0;
	Nop ();

	// Make IO as input and reading the BF
	TRISB = 0xFF;
	LCD_RS = 0;
	LCD_RW = 1;

	// Check for the Busy Flag	
	LCD_E = 1;
	Nop ();
	while (LCD_I == 1)
	{
		LCD_E = 0;
		Nop ();
		LCD_E = 1;
		Nop ();
	}

	#if defined (LCD_DEBUG_MODE)
	  DelayMs(2);
	#endif

	// Make IO as output
	LCD_E = 0;
	LCD_RW = 0;
	TRISB = 0x00;
}


/******************************************************************************
 * Function:        void LCDInit(void)
 * PreCondition:    None
 * Input:           None
 * Output:          None
 * Side Effects:    None
 *
 * Overview:        LCDText[] is blanked, port I/O pin TRIS registers are 
 *					configured, and the LCD is placed in the default state
 *
 * Note:            None
 *****************************************************************************/
void LCDInit(void)
{
	BYTE i;

	// Clear the LCD text buffer
	memset(LCDText, ' ', sizeof(LCDText)-1);
	LCDText[sizeof(LCDText)-1] = 0;

	// Setup the I/O pins
	LCD_E = 0;
	LCD_RW = 0;
	LCD_RS = 0;
	TRISB = 0x00;

	// Wait the required time for the LCD to reset
	DelayMs(40);

	// Set the default function
	// Go to 8-bit mode first to reset the instruction state machine
	// This is done in a loop 3 times to absolutely ensure that we get 
	// to 8-bit mode in case if the device was previously booted into 
	// 4-bit mode and our PIC got reset in the middle of the LCD 
	// receiving half (4-bits) of an 8-bit instruction
	LCD_RS = 0;
	PORTB = 0x03;
	Nop();						// Wait Data setup time (min 40ns)
	for(i = 0; i < 3u; i++)
	{
		LCD_E = 1;
		Delay10us(1);			// Wait E Pulse width time (min 230ns)
		LCD_E = 0;
		DelayMs(2);
	}
	
	// Use 8-bit mode with two lines
	PORTB = 0x38;
	LCD_E = 1;
	Delay10us(1);
	LCD_E = 0;
	DelayMs(2);

	// Set the entry mode
	// Increment after each write, do not shift
	PORTB = 0x06;
	LCD_E = 1;
	Delay10us(1);
	LCD_E = 0;
	DelayMs(2);

	// Set the display control
	// Turn display on, no cusor, no cursor blink
	PORTB = 0x0C;
	LCD_E = 1;
	Delay10us(1);
	LCD_E = 0;
	DelayMs(2);

	// Clear the display
	PORTB = 0x01;
	LCD_E = 1;
	Delay10us(1);
	LCD_E = 0;
	DelayMs(2);

/*
	// Set some custom graphic symbols defs
	LCDWrite(0, 0x40);
	// ASCII Symbol 1 - degree
	LCDWrite(1, 0b00000110);
	LCDWrite(1, 0b00001001);
	LCDWrite(1, 0b00001001);
	LCDWrite(1, 0b00000110);
	LCDWrite(1, 0b00000000);
	LCDWrite(1, 0b00000000);
	LCDWrite(1, 0b00000000);
	LCDWrite(1, 0b00000000);
	// ASCII Symbol 2 - <-|
	LCDWrite(1, 0b00000001);
	LCDWrite(1, 0b00000101);
	LCDWrite(1, 0b00001001);
	LCDWrite(1, 0b00011111);
	LCDWrite(1, 0b00001001);
	LCDWrite(1, 0b00000101);
	LCDWrite(1, 0b00000001);
	LCDWrite(1, 0b00000001);
*/
}


/******************************************************************************
 * Function:        void LCDUpdate(void)
 * PreCondition:    LCDInit() must have been called once
 * Input:           LCDText[]
 * Output:          None
 * Side Effects:    None
 *
 * Overview:        Copies the contents of the local LCDText[] array into the 
 *					LCD's internal display buffer.  Null terminators in 
 *					LCDText[] terminate the current line, so strings may be 
 *					printed directly to LCDText[].
 *
 * Note:            None
 *****************************************************************************/
void LCDUpdate(void)
{
	BYTE i, j;

	// ------------------ 1 x 16 LCD -----------------------------
	#if defined (USE_LCD_1x16_WINSTAR)
	// Go home for the first line
	LCDWrite(0, 0x80);
	for(i = 0; i < 8u; i++)
	{
		// Erase the rest of the line if a null char is 
		// encountered (good for printing strings directly)
		if(LCDText[i] == 0u)
			for(j=i; j < 8u; j++)
				LCDText[j] = ' ';
		LCDWrite(1, LCDText[i]);
	}
	// Set the address to the second line
	LCDWrite(0, 0xC0);
	for(i = 8; i < 16u; i++)
	{
		if(LCDText[i] == 0u)
			for(j=i; j < 16u; j++)
				LCDText[j] = ' ';
		LCDWrite(1, LCDText[i]);
	}
	#endif

	// ------------------ 1 x 16 LCD AMPIRE MODULE -----------------------------
	#if defined (USE_LCD_1x16_AMPIRE)
	// Go home for the first line
	LCDWrite(0, 0x00);
	for(i = 0; i < 8u; i++)
	{
		// Erase the rest of the line if a null char is 
		// encountered (good for printing strings directly)
		if(LCDText[i] == 0u)
			for(j=i; j < 8u; j++)
				LCDText[j] = ' ';
		LCDWrite(1, LCDText[i]);
	}
	for(i = 8; i < 40u; i++)
		LCDWrite(1, ' ');
	for(i = 8; i < 16u; i++)
	{
		if(LCDText[i] == 0u)
			for(j=i; j < 16u; j++)
				LCDText[j] = ' ';
		LCDWrite(1, LCDText[i]);
	}
	#endif

	// ------------------ 2 x 16 LCD -----------------------------
	#if defined (USE_LCD_2x16_WINSTAR)
	// Go home for the first line
	LCDWrite(0, 0x80);
	for(i = 0; i < 16u; i++)
	{
		// Erase the rest of the line if a null char is 
		// encountered (good for printing strings directly)
		if(LCDText[i] == 0u)
			for(j=i; j < 16u; j++)
				LCDText[j] = ' ';
		LCDWrite(1, LCDText[i]);
	}
	// Set the address to the second line
	LCDWrite(0, 0xC0);
	for(i = 16; i < 32u; i++)
	{
		if(LCDText[i] == 0u)
			for(j=i; j < 32u; j++)
				LCDText[j] = ' ';
		LCDWrite(1, LCDText[i]);
	}
	#endif

	// -------------- 4 x 20 LCD -------------------------
	#if defined (USE_LCD_4x20_WINSTAR)
	// Go home for the first line
	LCDWrite(0, 0x02);
	for(i = 0; i < 16u; i++)
	{
		// Erase the rest of the line if a null char is 
		// encountered (good for printing strings directly)
		if(LCDText[i] == 0u)
			for(j=i; j < 20u; j++)
				LCDText[j] = ' ';
		LCDWrite(1, LCDText[i]);
	}
	// Set the address to the second line
	LCDWrite(0, 0xC0);
	for(i = 20; i < 40u; i++)
	{
		if(LCDText[i] == 0u)
			for(j=i; j < 40u; j++)
				LCDText[j] = ' ';
		LCDWrite(1, LCDText[i]);
	}
	// Set the address to the third line
	LCDWrite(0, 0x94);
	for(i = 40; i < 60u; i++)
	{
		if(LCDText[i] == 0u)
			for(j=i; j < 60u; j++)
				LCDText[j] = ' ';
		LCDWrite(1, LCDText[i]);
	}
	// Set the address to the forth line
	LCDWrite(0, 0xD4);
	for(i = 60; i < 80u; i++)
	{
		if(LCDText[i] == 0u)
			for(j=i; j < 80u; j++)
				LCDText[j] = ' ';
		LCDWrite(1, LCDText[i]);
	}
	#endif
}

/******************************************************************************
 * Function:        void LCDErase(void)
 * PreCondition:    LCDInit() must have been called once
 * Input:           None
 * Output:          None
 * Side Effects:    None
 *
 * Overview:        Clears LCDText[] and the LCD's internal display buffer
 *
 * Note:            None
 *****************************************************************************/
void LCDErase(void)
{
	LCDWrite(0, 0x01);						// Clear display
	memset(LCDText, ' ', sizeof(LCDText));	// Clear local copy
}