/*********************************************************************
 *
 *     Time Function
 *
 *     A full description and usage examples for the functions in this
 *     file is located in time.txt
 *
 *********************************************************************
 * FileName:        time.c

 *
 * Software License Agreement
 *
 * This software is owned by Kenneth Lumia and is supplied to you 
 * for use exclusively as described in the associated software 
 * agreement.  This software is protected by software and other 
 * intellectual property laws.  Any use in violation of the software 
 * license may subject the user to criminal sanctions as well as 
 * civil liability.  Copyright 2006 Kenneth Lumia.  All rights reserved.
 *
 * This software is provided "AS IS."  KENNETH LUMIA DISCLAIMS ALL 
 * WARRANTIES, EXPRESS, IMPLIED, STATUTORY OR OTHERWISE, NOT LIMITED 
 * TO MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, AND 
 * INFRINGEMENT.  Kenneth Lumia shall in no event be liable for special, 
 * incidental, or consequential damages.
 *
 *
 * Author               Date   		Comment
 *~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 * Kenneth Lumia        10/02/2006  First Documented Version (time.txt)
 * Kenneth Lumia        10/04/2006  Updated to avoid potential changes 
 *                                  in system time tick (accessed both 
 *                                  by interrupt and mainline code)
 *                                  Archived as V100406a
 * Kenneth Lumia        11/26/2006  Release 1.0 - archived as V112606
 *~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 *                          ~~~~~~~~~~~~
 *                          Known Issues
 *                          ~~~~~~~~~~~~
 * 1. The leap year calculation is based soley on a 4 year cycle. Any code
 *    doing a leap year check will fail when the 400 year issue crops 
 *    up (year 2100).
 *
 * 2. Any routine that uses the yearday[] array will fail after 
 *    approximately year 2041 (indexing off the end of the table).
 *    Extending the table is a simple solution to push off the 
 *    eventual failure.
 *
 * 3. In February of 2036, the system_time variable will roll over
 *    reminicent of the Y2K bug.  This is not taken into account
 *    in the current code.  SNTP as per RFC4330 also has this issue
 *    which will need to be addressed.  This is common on many systems
 *    in use today and will cause the same grief as Y2K.
 *
 * 
 ********************************************************************/
#include "time.h"
#include "TCPIP Stack/TCPIP.h"
#include "TCPIP Stack/sntp.h"
#include "TCPIP Stack/helpers.h"
 
//Yearday is defined as the starting day(Jan 1) of each year, 
//starting at 0, i.e. 1/1/2006 = 0, 1/1/2007 = 365, etc.
rom int yearday[]= 			{0,365,730,1096,1461,1826,2191,2557,2922,3287,3652,4018,4383,4748,5113,5479,5844,6209,6574,6940,7305,7670,8035,8401,8766,9131,9496,9862,10227,10592,10957,11323,11688,12053,12418,12784};
rom int monthday[]= 		{-1,30,58,89,119,150,180,211,242,272,303,333,364};
rom int leapmonthday[] = 	{-1,30,59,90,120,151,181,212,243,273,304,334,365};
rom int keyoffset[] = 		{1,4,4,0,2,5,0,3,6,1,4,6};
rom int secondsunday[] = 	{9,8,14,13,12,11,10};
rom int firstsunday[] = 	{2,1,7,6,5,4,3};
rom char monthtext[] = 		"JanFebMarAprMayJunJulAugSepOctNovDec";
rom char dayofweektext[] = 	"SunMonTueWedThuFriSat";

int TMZone;            //standard hours from UTC, not including savings time
int DaylightBias;      //hours to adjust for DST (typically +1)

tm tm_MyTime;          //static buffer for a tm structure for use with gmtime and localtime
char asctimebuf[26];   //static buffer with ASCII date 	

/*********************************************************************
* Function:       Init_Time
*
* This function initializes the time variables. 
*
* Syntax:
* void Init_Time(void)
*
* Parameters:
* None
*
* Return Values:
* None
*
* PreCondition:    
* None
*
* Global Variables Modified:
* TMZone,DaylightBias
*
* Remarks:
* This should be called in the mainline code before any time 
* functions are used.
*********************************************************************/
void Init_Time(int TZ,int DL)  
{
   TMZone = TZ;                //force to EST for now
   DaylightBias = DL;          //force DST to 1 hour ahead
}

// Return the current time zone
int GetTimeZone(void)  
{
   return TMZone;                //force to EST for now
}

/*********************************************************************
* Function:       dayofweek(tm* tmptr)
*
* This function determines the day of week that is represented by
* the input tm structure.  
*
* Syntax:
* UINT_8 dayofweek(tm* tmptr)
*
* Parameters: pointer to a tm structure
*
* Return Values:
* Day of week: Sat=0, Sun=1, Mon=2, etc.
*
* PreCondition:  
* The following fields in the input structure are the only
* ones used for the calculation and must be correct:
*     tm_year, tm_mon, tm_mday
*
* Global Variables Modified:  None
* 
* Remarks
*
*********************************************************************/
BYTE dayofweek(tm* pp)
{
   //Calculation based on info from www.faqs.org/faqs/sci-math-faq/dayWeek/
   //This routine is only good for 2000-2099 dates.

   int yr_digits;
   int temp;

   //get last 2 digits of current year
   yr_digits = pp->tm_year - 100;                //106-100=06 for 2006

   temp = yr_digits/4;                           //integer division

   temp += pp->tm_mday;                          //add in the day of the month

   temp += keyoffset[pp->tm_mon];                //add in magic offset

   //determine if it is a leap year, taking advantage of integer division in calculation

   if((pp->tm_year - ((pp->tm_year/4)*4)) == 0)  //This is a leap year
   {  
       if(pp->tm_mon < 2)                        //jan or feb
          temp -= 1;                             //subtract 1 day

   }
 
   temp += 6;                                    //add 6 days for 2000 century dates and Gregorian calander

   temp += yr_digits;                            //add the year back in

   temp = temp - ((temp/7)*7);                   //get the remainder (Sat=0, Sun=1, Mon=2, etc.)

   return temp;
}

/*********************************************************************
* Function:       mktime(tm* tmptr)
*
* This function creates a time_t (32 bit) number based on the parts of
* the input tm structure.  The input fields in the tm structure are allowed
* to be out of there normal range (as long as they themselves don't overflow).
* This routine does not adjust for timezones or DST.
*
* Syntax:
* time_t mktime(tm* tmptr)
*
* Parameters: pointer to a tm structure
*
* Return Values:
* a time_t value (32bits)
*
* PreCondition:  
* The following fields in the input structure are the only
* ones used for the calculation and must be correct:
*     tm_year, tm_mon, tm_mday, tm_hour, tm_min, tm_sec
*
* Global Variables Modified:  None
* 
* Remarks
*
*********************************************************************/
DWORD mktime(tm* pp)   //convert a tm structure into a 32bit number
{
   int index;
   DWORD TimeTemp;  
   DWORD days;

   //TimeTemp = 3345062400;                        //start off at 1/1/2006 00:00:00
   TimeTemp = 1136073600;                          //start off at 1/1/2006 00:00:00  

   //determine the number of days since 1/1/2006 and current date

   index = pp->tm_year - 106;                    //calculate offset from 2006, (2006 = 0)

   days = (DWORD)(yearday[index]);              //add in days up to start of current year

   //calculate number of days into the current year (determine leap year first)
 
   if((pp->tm_year - ((pp->tm_year/4)*4)) != 0)  //not leap year,use regular year lookup
      days += (DWORD)monthday[pp->tm_mon];      //previous days up to this month
   else                                          //use leapyear lookup
      days += (DWORD)leapmonthday[pp->tm_mon];  //previous days up to this month

   days += pp->tm_mday;                          //add in the days in this month

   TimeTemp += days * 86400;                     //convert days to seconds

   TimeTemp += (DWORD)(pp->tm_hour) * 3600 ;    //add in the hours
   TimeTemp += (DWORD)(pp->tm_min)  * 60   ;    //add in the minutes
   TimeTemp += (DWORD)(pp->tm_sec)         ;    //add in the seconds

   return TimeTemp;
}

/*********************************************************************
* Function:       gmtime(const time_t *tmptr)
*
* This function expands the 32 bit time pointed to by tmptr into a 
* static tm structure that is the equivalent Universal Coordinated Time
* (UTC).  The buffer that is pointed to by the returned pointer is shared
* with any data from calls to the localtime() function.
*
* Syntax:
* tm* gmtime(const time_t *tmptr)
*
* Parameters: ptr to a timestamp
*
* Return Values:
* pointer to a structure that contains current date and time (UTC)
*
* PreCondition:  None
*
* Global Variables Modified:  tm_MyTime
* 
* Remarks
*
*********************************************************************/
tm* gmtime(const DWORD *TimeStamp)   // convert the timestamp into month, day, year, h,m,s
{
   int index;
   DWORD TimeRef;  
   DWORD days;
   DWORD hours;
   DWORD minutes;
   DWORD seconds;
   DWORD temp;
   tm* pp;


   pp = &tm_MyTime;                          //point to static variable

   TimeRef = *TimeStamp - 1136073600 ;     

   //  TimeRef = *TimeStamp - 3345062400;        //convert to time relative to 1/1/2006 00:00:00 
   // UTCSeconds =  UTCSeconds + (3345062400-1136073600); 


   //This must be converted into days, hours, minutes and seconds next.

   days = TimeRef/86400;                     //days integer division! 
   temp = TimeRef-(days*86400);              //whats left after days removed
   hours = temp/3600;                        //hours integer division!
   temp = temp -(hours*3600);                //whats left after hours removed
   minutes = temp/60;                        //minutes integer division!
   seconds = temp -(minutes*60);             //seconds

   //save to the structure (eventually combine with above statements to reduce code)
   pp->tm_sec = (int)seconds;              //save seconds
   pp->tm_min = (int)minutes;              //save minutes
   pp->tm_hour= (int)hours;                //save hours
   

   //still need year, month, day of month

   
   index = 0;  //Index through the table to determine the correct range

   while((yearday[index]) <= days)           //loop until we pass the correct number of days
       index++;                              //index will always be at least one here

   index--;                                  //adjust index into table to point to current year

   //index now holds the year offset.
   pp->tm_year = 2006 + index;               //save off the current year, base year = 2006 (non-standard)

   pp->tm_yday = (int)(days - yearday[index]);    //save number of days into the current year (0= Jan 1)

   //need to convert days into this year into a month and a day of month.

   index = 0;

   //determine if it is a leap year, taking advantage of integer division in calculation

   if((pp->tm_year - ((pp->tm_year/4)*4)) != 0)       //use regular year lookup
   {  
         while (monthday[index] < pp->tm_yday )       //calculate month
             index++;
       
         index--;                            //adjust index into table to point to current month (0-11)

         pp->tm_mday = pp->tm_yday - monthday[index]; //save off day of month (1-31)
   }
   else                                      //use leapyear lookup
   {  
        while (leapmonthday[index] < pp->tm_yday)    //calculate month
             index++;

        index--;        //adjust index into table to point to current month (0-11)

        pp->tm_mday = pp->tm_yday - leapmonthday[index]; //save off day of month (1-31)
   }

   pp->tm_mon = index;                       //save off the month (0-11)

   pp->tm_year = pp->tm_year - 1900;  		//adjust for standard offset from 1900

   pp->tm_wday = dayofweek(&tm_MyTime);      //determine what weekday the current date falls on 0=Sat, 1=Sun

   //correct so that Sunday=0, Mon = 1, Sat = 6
   pp->tm_wday -= 1;                         //first, adjust returned value by one day

   if (pp->tm_wday < 0)                      //did 0 -> -1 (Sat rollover)
      pp->tm_wday = 6;                       //roll Sat back around to 6


   return pp;
}

/*********************************************************************
* Function:       localtime(const time_t *tmptr)
*
* This function expands the 32 bit (UTC) time pointed to by tmptr into a 
* static tm structure that is converted to the local time based on timezone
* and daylight savings time (if active). The buffer that is pointed to by 
* the returned pointer is shared with any data from calls to the gmtime() 
* function.
*
* Syntax:
* tm* localtime(const time_t *tmptr)
*
* Parameters: ptr to a UTC timestamp
*
* Return Values:
* pointer to a structure that contains current date and time adjusted to
* local timezone and DST
*
* PreCondition:  None
*
* Global Variables Modified:  tm_MyTime
* 
* Remarks
*
*********************************************************************/
tm* localtime(const DWORD *TimeStamp)   // convert the UTC into local time
{
   DWORD TimeTemp;  
   tm* pp;
   DWORD offset;
   
   TimeTemp = *TimeStamp;

   //adjust for timezone
   offset = (DWORD)TMZone * 3600;      //convert hours offset into seconds   
   TimeTemp += offset;                 //adjust current time to UTC+TimeZone
   pp = gmtime(&TimeTemp);             // convert the timestamp into month, day, year, h,m,s
   
   return pp;
}


/*********************************************************************
* Function:       asctime
*
* This function converts the time in the tm input structure 
* into a string and stores it in a character buffer. 
*
* Syntax:
* char* asctime(tm * the_time)
*
* Parameters:
* a pointer to a time structure
*
* Return Values:
* a pointer to a static buffer, shared with ctime
*
* The format is always "Ddd Mmm DD HH:MM:SS YYYY\n" , so it is 
* always 25 characters, plus a terminating null, 
* Ddd = Mon, Tue, Wed, etc.
* Mmm = Jan, Feb, etc.
* DD = day (always 2 characters 0-31)
* HH = hour (always 2 characters 0-23)
* MM = minute (always 2 characters 0-59)
* SS = second (always 2 characters 0-59)
*
* PreCondition:    
* None
*
* Global Variables Modified
* asctimebuf[]
*
* Remarks
*
*********************************************************************/
char* asctime(tm *the_time)
{

    char* ptr;
    int lookupindex;
    int writeindex;
    int i;

    char buf[6];

    ptr = &asctimebuf[0];

    //**************************************
    //  byte index into the asctimebuffer
    //  Ddd Mmm DD HH:MM:SS YYYY\n
    //  0123456789012345678901234
    //            1         2
    //**************************************

    asctimebuf[2] = ':';      //put in the blank spaces
    asctimebuf[5] = ':';      //put in the blank spaces
    asctimebuf[8] = ' ';      //put in the blank spaces
    asctimebuf[9] = '\n';     //put in the :
    asctimebuf[10] = 0;       //put in the blank spaces

    //print out the hour
    writeindex = 0;                            //point to hour location start in buffer

    if(the_time->tm_hour < 10)                  //requires one digit, plus a leading fill char
    {
       asctimebuf[writeindex++]= '0';           //put in the leading zero, and bump the index
       asctimebuf[writeindex] =  the_time->tm_hour + '0';  //convert single digit to ascii
    }
    else                                        //multidigit 10-23
    {
       itoa(the_time->tm_hour,buf);             //convert to a string
       asctimebuf[writeindex++]= buf[0];        //copy 1st digit
       asctimebuf[writeindex]= buf[1];          //copy 2nd digit, disgard the rest of the buf string
    }

    //print out the min
    writeindex = 3;                            //point to minute location start in buffer


    if(the_time->tm_min < 10)                   //requires one digit, plus a leading fill char
    {
       asctimebuf[writeindex++]= '0';           //put in the leading zero, and bump the index
       asctimebuf[writeindex] =  the_time->tm_min + '0';  //convert single digit to ascii
    }
    else                                        //multidigit 10-59
    {
      itoa(the_time->tm_min,buf);               //convert to a string
      asctimebuf[writeindex++]= buf[0];         //copy 1st digit
      asctimebuf[writeindex]= buf[1];           //copy 2nd digit, disgard the rest of the buf string
    }

    //print out the second
    writeindex = 6;                            //point to second location start in buffer

    if(the_time->tm_sec < 10)                   //requires one digit, plus a leading fill char
    {
       asctimebuf[writeindex++]= '0';           //put in the leading zero, and bump the index
       asctimebuf[writeindex] =  the_time->tm_sec + '0';  //convert single digit to ascii
    }
    else                                        //multidigit 10-59
    {
       itoa(the_time->tm_sec,buf);              //convert to a string
       asctimebuf[writeindex++]= buf[0];        //copy 1st digit
       asctimebuf[writeindex]= buf[1];          //copy 2nd digit, disgard the rest of the buf string
    }
    return ptr;
} 


/*********************************************************************
* Function:       ascdate
*
* This function converts the time in the tm input structure 
* into a string and stores it in a character buffer. 
*
* Syntax:
* char* asctime(tm * the_time)
*
* Parameters:
* a pointer to a time structure
*
* Return Values:
* a pointer to a static buffer, shared with ctime
*
* The format is always "Ddd Mmm DD HH:MM:SS YYYY\n" , so it is 
* always 25 characters, plus a terminating null, 
* Ddd = Mon, Tue, Wed, etc.
* Mmm = Jan, Feb, etc.
* DD = day (always 2 characters 0-31)
* HH = hour (always 2 characters 0-23)
* MM = minute (always 2 characters 0-59)
* SS = second (always 2 characters 0-59)
*
* PreCondition:    
* None
*
* Global Variables Modified
* asctimebuf[]
*
* Remarks
*
*********************************************************************/
char* ascdate(tm *the_time)
{
    char* ptr;
    int lookupindex;
    int writeindex;
    int i;

    char buf[6];

    ptr = &asctimebuf[0];

    //**************************************
    //  byte index into the asctimebuffer
    //  Ddd Mmm DD HH:MM:SS YYYY\n
    //  0123456789012345678901234
    //            1         2
    //**************************************
    asctimebuf[3] = ' ';      //put in the blank spaces
    asctimebuf[6] = ' ';      //put in the blank spaces
    asctimebuf[11] = ',';     //put in the blank spaces
    asctimebuf[12] = ' ';     //put in the :
    asctimebuf[16] = ' ';     //put in the blank spaces
    asctimebuf[17] = '\n';    //put in the linefeed
    asctimebuf[18] = 0;       //put in the ending null

    //convert and enter the month data
    lookupindex = (the_time->tm_mon)*3;         //calc index into char array
    writeindex = 0;                             //starting location of month characters

    for(i=lookupindex;i<=lookupindex+2;i++)     //copy over 3 characters
      asctimebuf[writeindex++] = monthtext[i];  //copy the character and bump the indexes


    //print out the day
    writeindex = 4;                             //point to day location start in buffer
  
    if(the_time->tm_mday < 10)                  //requires one digit, plus a leading fill char
    {
       asctimebuf[writeindex++]= '0';           //put in the leading zero, and bump the index
       asctimebuf[writeindex] =  the_time->tm_mday + '0';  //convert single digit to ascii
    }
    else                                        //multidigit 10-31
    {
       itoa(the_time->tm_mday,buf);             //convert to a string
       asctimebuf[writeindex++]= buf[0];        //copy 1st digit
       asctimebuf[writeindex]= buf[1];          //copy 2nd digit, disgard the rest of the buf string
    }

    //print out the year
    writeindex = 7;                            //point to year location start in buffer

    itoa((1900 + the_time->tm_year) ,buf);      //adjust year from 1900 base, convert to ascii

    for(i=0;i<4;i++)                            //stringcpy year to output buffer
        asctimebuf[writeindex++]= buf[i];
   

    // convert the day of week
    lookupindex = (the_time->tm_wday)*3;        //calc index into char array
    writeindex = 13;                             //starting location of Day characters

    for(i=lookupindex;i<=lookupindex+2;i++)     //copy over 3 characters
       asctimebuf[writeindex++] = dayofweektext[i];  //copy the character and bump the indexes

    return ptr;
} 

/*********************************************************************
* Function:       ascdate
*
* This function converts the time in the tm input structure 
* into a string and stores it in a character buffer. 
*
* Syntax:
* char* asctime(tm * the_time)
*
* Parameters:
* a pointer to a time structure
*
* Return Values:
* a pointer to a static buffer, shared with ctime
*
* The format is always "Ddd Mmm DD HH:MM:SS YYYY\n" , so it is 
* always 25 characters, plus a terminating null, 
* Ddd = Mon, Tue, Wed, etc.
* Mmm = Jan, Feb, etc.
* DD = day (always 2 characters 0-31)
* HH = hour (always 2 characters 0-23)
* MM = minute (always 2 characters 0-59)
* SS = second (always 2 characters 0-59)
*
* PreCondition:    
* None
*
* Global Variables Modified
* asctimebuf[]
*
* Remarks
*
*********************************************************************/
char* ascdate1(tm *the_time)
{
    char* ptr;
    int lookupindex;
    int writeindex;
    int i;

    char buf[6];

    ptr = &asctimebuf[0];

    //**************************************
    //  byte index into the asctimebuffer
    //  Ddd Mmm DD HH:MM:SS YYYY\n
    //  0123456789012345678901234
    //            1         2
    //**************************************
  //  asctimebuf[3] = ' ';      //put in the blank spaces
  //  asctimebuf[6] = ' ';      //put in the blank spaces
  //  asctimebuf[11] = ',';     //put in the blank spaces
  //  asctimebuf[12] = ' ';     //put in the :
  //  asctimebuf[16] = ' ';     //put in the blank spaces
  //  asctimebuf[17] = '\n';    //put in the linefeed
    asctimebuf[9] = 0;       //put in the ending null

    //convert and enter the month data
    lookupindex = (the_time->tm_mon)*3;         //calc index into char array
    writeindex = 2;                             //starting location of month characters

    for(i=lookupindex;i<=lookupindex+2;i++)     //copy over 3 characters
      asctimebuf[writeindex++] = monthtext[i];  //copy the character and bump the indexes


    //print out the day
    writeindex = 0;                             //point to day location start in buffer
  
    if(the_time->tm_mday < 10)                  //requires one digit, plus a leading fill char
    {
       asctimebuf[writeindex++]= '0';           //put in the leading zero, and bump the index
       asctimebuf[writeindex] =  the_time->tm_mday + '0';  //convert single digit to ascii
    }
    else                                        //multidigit 10-31
    {
       itoa(the_time->tm_mday,buf);             //convert to a string
       asctimebuf[writeindex++]= buf[0];        //copy 1st digit
       asctimebuf[writeindex]= buf[1];          //copy 2nd digit, disgard the rest of the buf string
    }

    //print out the year
    writeindex = 5;                            //point to year location start in buffer

    itoa((1900 + the_time->tm_year) ,buf);      //adjust year from 1900 base, convert to ascii

    for(i=0;i<4;i++)                            //stringcpy year to output buffer
        asctimebuf[writeindex++]= buf[i];
   

    // convert the day of week
    //lookupindex = (the_time->tm_wday)*3;        //calc index into char array
    //writeindex = 13;                             //starting location of Day characters
//
//    for(i=lookupindex;i<=lookupindex+2;i++)     //copy over 3 characters
//       asctimebuf[writeindex++] = dayofweektext[i];  //copy the character and bump the indexes

    return ptr;
} 


/*********************************************************************
* Function:       ctime
*
* This function converts the date and time pointed to by timeptr to
* localtime and formats the time as a string.
*
* Syntax:
* char* ctime(time_t* timeptr)
*
* Parameters:
* a pointer to a 32-bit time value
*
* Return Values:
* a pointer to a static buffer, shared with asctime
*
* The format is always "Ddd Mmm DD HH:MM:SS YYYY\n" , so it is 
* always 25 characters, plus a terminating null, 
* Ddd = Mon, Tue, Wed, etc.
* Mmm = Jan, Feb, etc.
* DD = day (always 2 characters 0-31)
* HH = hour (always 2 characters 0-23)
* MM = minute (always 2 characters 0-59)
* SS = second (always 2 characters 0-59)
*
* PreCondition:    
* None
*
* Global Variables Modified
* asctimebuf[]
*
* Remarks
*
*********************************************************************/
char* ctime(DWORD *timeptr)
{
    char* ptr;
    ptr = asctime(localtime(timeptr));          //convert number to a time structure and then to a string   
    return ptr;
} 

char* cdate(DWORD *timeptr)
{
    char* ptr;
    ptr = ascdate(localtime(timeptr));          //convert number to a time structure and then to a string   
    return ptr;
}

char* pktdate(DWORD *timeptr)
{
    char* ptr;
    ptr = ascdate1(localtime(timeptr));          //convert number to a time structure and then to a string   
    return ptr;
}